'use strict';
/**
 *  @classdesc
 *  Memespeech hides encrypted text inside other text (Carrier Text) by
 *  converting the Carrier Text iNtO MEmEspEeCH. dON't LEt yoUR MeMES Be drEAmS!
 *
 *  @author     henriquez <henriquez@protonmail.com>
 *  @copyright  Copyright (c) 2019, henriquez.
 *              ({@link https://www.obsessivefacts.com})
 *  @license    {@link https://opensource.org/licenses/GPL-3.0|GPLv3}
 *
 *  @param {Object} options - A list of properties used to initialize the class.
 *
 *  @param {string} [options.carrierText] - Carrier Text we will hide the
 *      Cryptographic Payload inside, converting to Memespeech in the process.
 *
 *  @param {number} [options.pbkdf2Iterations=50000] - Number of iterations to
 *      use for key derivation. Must be a positive integer multiple of 10,000
 *      (maximum 150,000). Higher number means more secure but slower keygen.
 *
 *  @param {boolean} [options.debugMode] - Enables verbose console output of
 *      various internal workings.
 *
 *  @param {string} [options.workerUrl] - URL or path (relative to the root) for
 *      Memespeech Web Worker, auto-invokes {@link Memespeech#initializeWorker}.
 *      NOTE: the Web Worker is not supported under Node.js. Don't even try it!
 *
 *  @returns {Memespeech}
 */
let Memespeech = function(options) {
    if (!options) options = {}

    if (options.debugMode)
        this.setDebugModeActive(options.debugMode);

    if (options.workerUrl)
        this.initializeWorker(options.workerUrl);

    if (options.carrierText)
        this.setCarrierText(options.carrierText);

    if (options.pbkdf2Iterations)
        this.setPBKDF2Iterations(options.pbkdf2Iterations);

    return this;
}

// Detect if running under Node.js and do a little bit of housekeeping if so
if (typeof module !== 'undefined' && module.exports) {
    var crypto = require('crypto'),
        isNodeJS = true;
        
    module.exports = Memespeech;
}

/**
 *  The Carrier Text we will be hiding ciphertext inside by re-encoding as
 *  Memespeech. You can add this in as an option to the {@link Memespeech} init,
 *  or use {@link Memespeech#setCarrierText} to set it manually. NOTE: Do not
 *  write directly to this property or you will be punished.
 *  @type {string}
 *  @private
 */
Memespeech.prototype.carrierText = '';

/**
 *  The budget in bits for any Cryptographic Payload we may wish to encode into
 *  the Carrier Text. Auto-populated by {@link Memespeech#setCarrierText}.
 *  @type {number}
 *  @private
 */
Memespeech.prototype.carrierTextBudget = 0;

/**
 *  The array of alphabetic characters (and their positions) matched inside the
 *  {@link Memespeech#carrierText}. This is populated automatically by
 *  {@link Memespeech#setCarrierText} and allows us to later re-encode the
 *  Carrier Text as Memespeech, where the capitalization of each letter
 *  represents the binary bits (one or zero) inside the Cryptographic Payload.
 *  @type {Array}
 *  @private
 */
Memespeech.prototype.carrierTextABCMatches = [];

/**
 *  The key used for AES encryption and decryption. Populated by
 *  {@link Memespeech#initCipherFromPassphrase}
 *  @type {CryptoKey}
 *  @private
 */
Memespeech.prototype.blockCipherKey = null;

/**
 *  Initial Vector used for AES encryption and decryption. Populated by
 *  {@link Memespeech#initCipherFromPassphrase}
 *  @type {Uint8Array}
 *  @private
 */
Memespeech.prototype.blockCipherIV = null;

/**
 *  Ciphertext generated by {@link Memespeech#encrypt}
 *  @type {Uint8Array}
 *  @private
 */
Memespeech.prototype.ciphertext = null;

/**
 *  The Encoding Mode to use for the Cryptographic Payload. May be either
 *  'standard' or 'compact'. DO NOT SET THIS DIRECTLY. This can overridden in
 *  {@link Memespeech#initCipherFromPassphrase}. Standard Mode uses a 32 byte
 *  salt and a signature verification on decryption. Compact Mode significantly
 *  reduces payload size with an 8 byte salt and no verification on decryption.
 *
 *  @type {string}
 *  @private
 */
Memespeech.prototype.encodingMode = 'standard';

/**
 *  Salt bytes used for PBKDF2 key derivation. Needs to be the correct length
 *  for the selected {@link Memespeech#encodingMode} and is typically generated 
 *  by {@link Memespeech#initCipherFromPassphrase}. Generally this should not be
 *  set directly, unless you are very clever.
 *  @type {Uint8Array}
 *  @private
 */
Memespeech.prototype.pbkdf2Salt = null;

/**
 *  Number of iterations used for PBKDF2 key derivation. This must be a positive
 *  integer multiple of 10,000 (maximum 150,000). This can be set via an option
 *  to {@link Memespeech} instantiation, or set prior to key derivation by
 *  calling {@link Memespeech#setPBKDF2Iterations}.
 *  @type {number}
 *  @private
 */
Memespeech.prototype.pbkdf2Iterations = 50000;

/**
 *  Array of binary "bits" (integer zeroes and ones) representing our
 *  Cryptographic Payload. This is populated by {@link Memespeech#encrypt}.
 *  @type {Array}
 *  @private
 */
Memespeech.prototype.cryptographicPayload = null;

/**
 *  Whether to enable debug mode (verbose console output) for this instance.
 *  Add this as an option to the {@link Memespeech} instantiation function,
 *  or call {@link Memespeech#setDebugModeActive} to turn it on.
 *  @type {boolean}
 *  @private
 */
Memespeech.prototype.debugMode = false;

/**
 *  Whether the instance is running in a Node JS context. This will slightly
 *  affect some crypto-specific functionality.
 *  @type {boolean}
 *  @private
 */
Memespeech.prototype.isNodeJS = typeof isNodeJS !== "undefined";

/**
 *  The Memespeech Web Worker (if initialized, see
 *  {@link Memespeech#initializeWorker}).
 *  @type {Worker}
 *  @private
 */
Memespeech.prototype.worker = null;

/**
 *  Holds "callback" functions to resolve async Promises that use the Worker.
 *  The defaults here get overridden by the relevant instance methods as needed.
 *  @type {Object}
 *  @private
 */
Memespeech.prototype.workerCallbacks = {
    analyzeCarrierText: {
        resolve: null,
        reject: null
    },
    computeCryptographicPayload: {
        resolve: null,
        reject: null
    },
    toString: {
        resolve: null,
        reject: null
    },
    parse: {
        resolve: null,
        reject: null
    }
};

/**
 *  Initializes the Memespeech Web Worker and binds its communication events.
 *  The worker allows you to optionally run computationally-expensive processes
 *  on a separate CPU thread, so as not to slow down browser UI performance
 *  while processing data.
 *
 *  Please check your JavaScript console for any 404 warnings when calling this.
 *  Web Workers have an annoying habit of silently failing on 404. Also note
 *  that none of the Worker stuff is supported under Node.js.
 *
 *  @param {String} workerUrl - URL or path (relative to root) of the worker lib
 *  @returns {Memespeech}
 */
Memespeech.prototype.initializeWorker = function(workerUrl) {
    let id = '(Memespeech Worker Handler)',
        worker = new Worker(workerUrl),
        cb = this.workerCallbacks;

    worker.addEventListener('message', function(e) {
        switch (e.data.call) {
            case 'debugWrite':
                this.debugWrite(' ~ worker ~ ', e.data.message);
                break;
            case 'analyzeCarrierText:resolve':
                cb.analyzeCarrierText.resolve(e.data.matches);
                break;
            case 'analyzeCarrierText:reject':
                cb.analyzeCarrierText.reject(e.data.err);
                break;
            case 'computeCryptographicPayload:resolve':
                cb.computeCryptographicPayload.resolve(e.data.payload);
                break;
            case 'computeCryptographicPayload:reject':
                cb.computeCryptographicPayload.reject(e.data.err);
                break;
            case 'toString:resolve':
                cb.toString.resolve(e.data.memespeechStr);
                break;
            case 'toString:reject':
                cb.toString.reject(e.data.err);
                break;
            case 'parse:resolve':
                cb.parse.resolve(e.data);
                break;
            case 'parse:reject':
                cb.parse.reject(e.data.err);
                break;
        }
    }.bind(this), false);

    worker.postMessage({ call: 'sayHello' });
    this.worker = worker;

    return this;
}

/**
 *  Sets the Carrier Text & analyzes to set {@link Memespeech#carrierTextBudget}
 *  and {@link Memespeech#carrierTextABCMatches}. We can use these later on to
 *  assist the process of computing the Cryptographic Payload and encoding it
 *  into Memespeech text.
 *
 *  @param {string} text - A string
 *  @returns {Memespeech}  This instance
 */
Memespeech.prototype.setCarrierText = function(text) {
    this.carrierText = text;
    let start = new Date().getTime(),
        matches = [...this.carrierText.matchAll(/[A-Za-z]/g)],
        end = new Date().getTime();

    this.carrierTextBudget = matches.length;
    this.carrierTextABCMatches = matches;

    this.debugWrite('(Memespeech.analyzeCarrierText)')
    this.debugWrite('    carrier ABC budget:  ', matches.length)
    this.debugWrite('    elapsed time (ms):   ', (end - start))
    return this;
}

/**
 *  Async version of {@link Memespeech#setCarrierText}. Sets the Carrier Text
 *  and calls out to the Memespeech Web Worker to analyze the text for
 *  replaceable alphabetic characters. Use this if the synchronous version is
 *  clogging your main thread. NOTE: the Worker has overhead, so using this can
 *  take longer, but if you're counting FPS it can be worth using this for
 *  better "perceived" UI performance.
 *
 *  This method is not supported under Node.js.
 *
 *  @param {string} text - A string
 *  @returns {Promise}     A promise that will resolve when the work is done.
 */
Memespeech.prototype.setCarrierTextAsync = async function(text) {
    let id = '(Memespeech.setCarrierTextAsync)',
        cb = this.workerCallbacks.analyzeCarrierText,
        start = new Date().getTime()

    if (!this.worker)
        throw id + ': Worker not initialized. Pls call initializeWorker first!';

    this.carrierText = text;

    return new Promise(function(resolve, reject) {
        cb.resolve = function(matches) {
            let end = new Date().getTime();

            this.carrierTextBudget = matches.length;
            this.carrierTextABCMatches = matches;

            this.debugWrite(id + ' -- Received from Worker!');
            this.debugWrite('    carrier ABC budget:  ', matches.length);
            this.debugWrite('    elapsed time (ms):   ', (end - start));
            resolve(matches);
        }.bind(this);

        cb.reject = function(err) {
            console.error(id + ' Worker FAIL: ', err);
            reject(err);
        }.bind(this);

        this.worker.postMessage({
            call: 'analyzeCarrierText',
            text: text
        })
    }.bind(this))
}

/**
 *  Generates a random salt value. Depending on the Encoding Mode (see
 *  {@link Memespeech#encodingMode}) it will be either 32 bytes (standard) or 8
 *  bytes (compact). This is used by {@link Memespeech#initCipherFromPassphrase}
 *  to create a unique salt for this instance. The salt is stored in the Footer
 *  Packet of the Cryptographic Payload to prevent rainbow attacks against 
 *  ciphertext encrypted with the same passphrase.
 *
 *  @private
 *  @returns {Memespeech}
 */
Memespeech.prototype.generateRandomSalt = async function() {
    let byteLength = this.encodingMode == 'standard' ? 32 : 8;
    if (this.isNodeJS) {
        this.pbkdf2Salt = await Memespeech.DataUtils.nodeRandomBytes(byteLength)
    } else {
        this.pbkdf2Salt = crypto.getRandomValues(new Uint8Array(byteLength));
    }
    return this;
}

/**
 *  Sets the number of iterations for PBKDF2 key derivation. Higher numbers mean
 *  slower key generation but better security. This must be an integer multiple
 *  of 10,000 (max 150,000).
 *
 *  @param {number} iterations - Positive integer PBKDF2 iterations (max 150000)
 *  @returns {Memespeech}
 */
Memespeech.prototype.setPBKDF2Iterations = function(iterations) {
    iterations = parseInt(iterations);
    if (
        isNaN(iterations)
        ||
        iterations < 1
        ||
        iterations > 150000
        ||
        iterations % 10000 !== 0
    ) {
        throw '(Memespeech.setPBKDF2Iterations): iterations must be a positive '
            + 'integer multiple of 10,000 (maximum 150,000)';
    }
    this.pbkdf2Iterations = iterations;
    return this;
}

/**
 *  Async function to generate an AES symmetric *cryption key and initial vector
 *  using PBKDF2 over a SHA-384 hasher. This populates a random salt value
 *  ({@link Memespeech#pbkdf2Salt}) and hashes with the number of iterations
 *  specified in {@link Memespeech#pbkdf2Iterations}. Both of these properties
 *  are persisted through the instance and stored in the Cryptographic Payload
 *  to make decryption possible.
 *
 *  @param {string} passphrase - A string to use as passphrase (aka password)
 *
 *  @param {string} [encodingMode] - Optional override to the Encoding Mode. If
 *      not passed, the {@link Memespeech#encodingMode} already set on this
 *      instance is used. If passed, acceptable values are either 'standard' or
 *      'compact'. Standard Mode uses a 32 byte salt whereas Compact Mode
 *      reduces to 8 bytes, allowing encoding into smaller Carrier Text strings
 *      with slightly worse security.
 *
 *  @returns {Memespeech}
 */
Memespeech.prototype.initCipherFromPassphrase = async function(passphrase,
    encodingMode) {

    let id = '(Memespeech.initCipherFromPassphrase)';

    if (!passphrase)
        throw id + ': passphrase is required, silly.';   

    if (encodingMode) {
        if (encodingMode != 'standard' && encodingMode != 'compact') {
            throw id + ': encodingMode must be \'standard\' or \'compact\.'; 
        }
        this.encodingMode = encodingMode;
    } 
        
    let allowableSaltLength = this.encodingMode == 'standard' ? 32 : 8;

    if (!this.pbkdf2Salt || this.pbkdf2Salt.length != allowableSaltLength)
        await this.generateRandomSalt();

    let symKey, iv, end, start;

    start = new Date().getTime();

    if (this.isNodeJS) {
        let keyIV = await new Promise((resolve, reject) => {
            crypto.pbkdf2(
                Memespeech.DataUtils.stringToBytes(passphrase),
                this.pbkdf2Salt,
                this.pbkdf2Iterations,
                48,
                'sha384',
                (err, derivedKey) => {
                    resolve(new Uint8Array(derivedKey));
                }
            );
        })

        this.blockCipherKey = keyIV.slice(0, 32);
        this.blockCipherIV = keyIV.slice(32);
    } else {
        let importedKey = await crypto.subtle.importKey(
            'raw',
            Memespeech.DataUtils.stringToBytes(passphrase),
            'PBKDF2',
            false,
            ['deriveBits']
        ),
        keyIV = await crypto.subtle.deriveBits(
            {
                'name': 'PBKDF2',
                'salt': this.pbkdf2Salt,
                'iterations': this.pbkdf2Iterations,
                'hash': 'SHA-384'
            },
            importedKey,
            384
        ),
        key = keyIV.slice(0, 32);
        iv = keyIV.slice(32);
        symKey = await crypto.subtle.importKey(
            'raw',
            key,
            'AES-CBC',
            false,
            ['encrypt', 'decrypt']
        );

        this.blockCipherKey = symKey;
        this.blockCipherIV = iv;
    }
    end = new Date().getTime();

    this.debugWrite(id);
    this.debugWrite('    salt:              ', this.pbkdf2Salt);
    this.debugWrite('    iterations:        ', this.pbkdf2Iterations);
    this.debugWrite('    elapsed time (ms): ', (end - start));

    return this;
}

/**
 *  Encrypt plaintext using AES-CBC. The block cipher properties for this
 *  instance must be populated prior to using this, typically by calling
 *  {@link Memespeech#initCipherFromPassphrase}.
 *
 *  @param {string} plaintext - The plaintext string to encrypt
 *  @returns {Memespeech}       
 */
Memespeech.prototype.encrypt = async function(plaintext) {
    let id = '(Memespeech.encrypt)';

    if (!this.blockCipherKey || !this.blockCipherIV) {
        throw id + ': cipher is not initialized for this instance. You must '
            + 'first use initCipherFromPassphrase to populate key and IV data.';
    }
    let start = new Date().getTime(),
        encodedPlaintext = Memespeech.DataUtils.encodePlaintext(
            plaintext,
            this.encodingMode
        ),
        ciphertext;
    
    if (this.isNodeJS) {
        let cipher = crypto.createCipheriv(
            'aes-256-cbc',
            this.blockCipherKey,
            this.blockCipherIV
        );
        let encrypted = cipher.update(encodedPlaintext);
        ciphertext = new Uint8Array(Buffer.concat([encrypted, cipher.final()]));
    } else {
        ciphertext = new Uint8Array(await crypto.subtle.encrypt(
            { name: 'AES-CBC', iv: this.blockCipherIV },
            this.blockCipherKey,
            encodedPlaintext
        ));
    }
    let end = new Date().getTime();

    this.ciphertext = ciphertext;

    this.debugWrite(id);
    this.debugWrite('    ciphertext:        ', ciphertext);
    this.debugWrite('    elapsed time (ms): ', (end - start));

    if (this.worker) return await this.computeCryptographicPayloadAsync();

    return this.computeCryptographicPayload();
}

/**
 *  Convenience function to both derive an AES key from a passphrase and use it
 *  to encrypt plaintext.
 *
 *  Internally calls {@link Memespeech#initCipherFromPassphrase} followed by
 *  {@link Memespeech#encrypt}, returns a Promise that resolves to the instance.
 *
 *  @param {string} plaintext  - The plaintext string to encrypt
 *  @param {string} passphrase - The passphrase used for key derivation
 *  @param {string} [encodingMode] - Override to the Encoding Mode. See the
 *      description in {@link Memespeech#initCipherFromPassphrase} for more info
 *  @returns {Memespeech}       
 */
Memespeech.prototype.passphraseEncrypt = async function(plaintext, passphrase,
    encodingMode) {
    await this.initCipherFromPassphrase(passphrase, encodingMode);
    return await this.encrypt(plaintext);
}

/**
 *  Generates a fake Cryptographic Payload and encodes it into Memespeech text
 *  in the case we have nothing to encrypt but still want to generate Memespeech
 *  text. Does not actually encrypt anything. This provides plausible
 *  deniability for other people who _are_ encrypting, since it will be possible
 *  to parse decoy messages but impossible to decrypt. Is there something
 *  encrypted in there? Who knows! There might not be. Maybe this is all a big
 *  fucking joke.
 *
 *  The decoy Cryptographic Payload will use the strongest Encoding Mode
 *  available (based on the budget of the Carrier Text), with a randomized
 *  number of blocks of fake ciphertext. If the Carrier Text is too short to
 *  encode any Cryptographic Payload into, this function will pop a warning and
 *  simply randomize the alphabetical characters of the Carrier Text to simulate
 *  Memespeech.
 *
 *  Throws an error if no Carrier Text is present on this instance.
 *
 *  @returns {String}
 */
Memespeech.prototype.generateDecoyText = async function() {
    let id = '(Memespeech.createDecoy)',
        encodingMode,
        fakePassphrase,
        maxFakeBytes,
        fakeBytes;

    if (!this.carrierText) {
        throw id + ': No Carrier Text is set. Please set some and try again.';
    }

    // First, let's see what kind of budget we're working with. We will
    // simulate the "strongest" possible Encoding Mode for a Cryptographic
    // Payload that would fit within the Carrier Text (if any).
    let budgets = this.computeAESByteBudget();

    // Based on the budget, figure out what Encoding Mode to use and compute the
    // maximum number of bytes of ciphertext we could fit into our payload.
    if (budgets.byteBudgetStandard > 0) {
        encodingMode = 'standard';
        maxFakeBytes = budgets.byteBudgetStandard + 5;
    } else if (budgets.byteBudgetCompact > 0) {
        encodingMode = 'compact';
        maxFakeBytes = budgets.byteBudgetCompact + 1;
    } else {
        // if no Cryptographic Payload will fit into the Carrier Text, then
        // simply randomize the alphabetization of the bytes of Carrier Text.
        console.warn(id + ": Can't fit ANY ciphertext in the Carrier Text.")
        if (this.worker) return await this.toStringAsync();
        else return this.toString();
    }
    
    // Just because we could fit max bytes doesn't mean we should.
    // Randomly choose a byte length based on the size of 16 byte AES blocks.
    fakeBytes = (Math.floor(Math.random() * maxFakeBytes / 16) + 1) * 16;
    
    // Generate fake ciphertext and passphrase
    if (this.isNodeJS) {
        fakePassphrase = await Memespeech.DataUtils.nodeRandomBytes(16);
        this.ciphertext = await Memespeech.DataUtils.nodeRandomBytes(fakeBytes);
    } else {
        fakePassphrase = crypto.getRandomValues(new Uint8Array(16));
        this.ciphertext = crypto.getRandomValues(new Uint8Array(fakeBytes));
    }
    fakePassphrase = Memespeech.DataUtils.bytesToString(fakePassphrase);

    // Use the fake passphrase to initialize salt and IV for the cipher
    await this.initCipherFromPassphrase(fakePassphrase, encodingMode);

    this.debugWrite(id);
    this.debugWrite('    byte budgets:   ', budgets);
    this.debugWrite('    encodingMode:   ', encodingMode);
    this.debugWrite('    maxFakeBytes:   ', maxFakeBytes);
    this.debugWrite('    fakeBytes:      ', fakeBytes);
    this.debugWrite('    fakePassphrase: ', fakePassphrase);
    this.debugWrite('    ciphertext:     ', this.ciphertext);

    // Lastly turn the fake ciphertext, salt and IV into a Cryptographic Payload
    // and encode it into Memespeech text.
    if (this.worker) {
        await this.computeCryptographicPayloadAsync();
        return await this.toStringAsync();
    }
    this.computeCryptographicPayload();
    return this.toString();
}

/**
 *  Encodes the ciphertext, PBKDF2 salt and iteration count into a Cryptographic
 *  Payload, which is an array of integer bits (zeroes and ones) that can be
 *  encoded into the Carrier Text to generate Memespeech text. The Cryptographic
 *  Payload is split into a Ciphertext Packet, which contains information about
 *  the ciphertext, and a Footer Packet, which contains information about the
 *  encryption cipher. This function is called automatically by
 *  {@link Memespeech#encrypt} and sets {@link Memespeech#cryptographicPayload}
 *  for the instance.
 *  @private
 */
Memespeech.prototype.computeCryptographicPayload = function() {
    let id = '(Memespeech.computeCryptographicPayload)',
        ciphertext = this.ciphertext,
        ciphertextBitLength = this.ciphertext.length * 8,
        start = new Date().getTime();

    if (!ciphertext)
        throw id + ': ciphertext is missing. Please encrypt first!';

    /*
     * The bit budget equals the number of alphabetical characters within the
     * Carrier Text. Each of these characters can be made lowercase or uppercase
     * to represent bit values 0 or 1, respectively.
     */
    let bitBudget = this.carrierTextBudget;

    /*
     * The number of bits in the Footer Packet depends on the Encoding Mode
     * selected at the time of key derivation. There are two possible values for
     * this.
     *
     * STANDARD ENCODING MODE FOOTER PACKET:
     *
     *      2   : bits 0 and 1 specifying that this is Standard Mode
     *    + 256 : bit length of our salt (32 bytes in Standard Mode)
     *    + 4   : four bits compressing the PBKDF2 iteration count (max 150000)
     *    = 262
     *
     * COMPACT ENCODING MODE FOOTER PACKET:
     *
     *      1   : bit 1 specifying that this is Compact Mode
     *    + 64  : bit length of our salt (8 bytes in Compact Mode)
     *    + 4   : four bits compressing the PBKDF2 iteration count (max 150000)
     *    = 69 lol
     *
     * Note that in both cases the actual iteration count is divided by 10,000
     * and compressed into 4 bits (which is why we require an even multiple of
     * 10,000 for iterations with a max of 150,000)
     */
    let footerPacketBits = this.encodingMode == 'standard' ? 262 : 69;

    /*
     * The Ciphertext Packet is randomly divided into chunks consisting of:
     *
     *      [LENGTHBYTE] + [bits] + [isLastChunk]
     *
     * More precisely, each chunk starts with eight bits forming a length
     * byte indicating how many bits of ciphertext data this chunk contains
     * (between 1 and 255), followed by the actual ciphertext data bits,
     * followed by a boolean bit indicating whether this is the last chunk of
     * the Ciphertext Packet.
     *
     * We define ciphertextPacketBudget as the total available number of bits
     * left over in the bit budget after subtracting the Footer Packet bit size.
     */
    let ciphertextPacketBudget = bitBudget - footerPacketBits;

    /*
     * minDivisions is the worst-case scenario: the minimum number of chunks
     * required to fully encode the Ciphertext Packet assuming that all length
     * bytes take the maximum 255 value. This would not be ideal because the
     * Memespeech output would end up with at least 8 consecutive capitalized
     * letters, which would not look very random.
     */
    let minCiphertextChunks = (Math.floor(ciphertextBitLength / 255) + 1);
    
    /*
     * minCiphertextPacketSize is the bit length of the Ciphertext Packet in the
     * worst-case that we end up using the minimum possible number of chunks
     */
    let minCiphertextPacketSize = 9 * minCiphertextChunks + ciphertextBitLength;

    /*
     * Throw an error if there's simply no way to fit the Ciphertext Packet
     * within its budget.
     */
    if (minCiphertextPacketSize > ciphertextPacketBudget) {
        throw id +': Cryptographic Payload cannot fit within the Carrier Text. '
            + 'Please try again with larger Carrier Text or smaller ciphertext.'
    }

    /*
     * Now actually convert the ciphertext from a Uint8Array to an array of
     * "binary bits" (where each bit is represented by an integer 0 or 1)
     */
    let ciphertextBits = Memespeech.DataUtils.bytesToBits(ciphertext),
        budgetUsage = minCiphertextPacketSize,
        chunks = [],
        anchor = 0;

    /*
     * Split ciphertext into chunks of up to (but ideally less than) 255 bits,
     * being careful to stay within our budget.
     */
    for (var i=0; i<minCiphertextChunks-1; i++) {
        if (budgetUsage + 9 < ciphertextPacketBudget) {
            let slicePoint = Math.floor(Math.random() * 255) + 1;
            chunks.push(ciphertextBits.slice(anchor, anchor + slicePoint));
            chunks.push(ciphertextBits.slice(anchor + slicePoint, anchor + 255))
            budgetUsage += 9;
        } else {
            chunks.push(ciphertextBits.slice(anchor, anchor + 255));
        }
        anchor += 255;
    }
    // fallback setting for very small ciphertexts
    if (chunks.length == 0 && budgetUsage + 9 <= ciphertextPacketBudget) {
        let slicePoint = Math.floor(Math.random() * ciphertextBits.length) + 1;
        chunks.push(ciphertextBits.slice(0, slicePoint));
        chunks.push(ciphertextBits.slice(slicePoint, 255));
    } else {
        chunks.push(ciphertextBits.slice(anchor, anchor + 255));
    }

    /*
     * Now form our final Cryptographic Payload based on all these chunks,
     * plus the Footer Packet data.
     */
    let cryptographicPayload = [];

    // first construct the Ciphertext Packet
    for (let i=0; i<chunks.length; i++) {
        let lengthByte = Memespeech.DataUtils.intToBits(chunks[i].length);
        for (let bit of lengthByte) cryptographicPayload.push(bit);
        for (let bit of chunks[i]) cryptographicPayload.push(bit);
        cryptographicPayload.push(i == chunks.length - 1 ? 1 : 0);
    }

    // next add in the Footer Packet
    let saltBits = Memespeech.DataUtils.bytesToBits(this.pbkdf2Salt),
        iterationsBits = Memespeech.DataUtils.byteToBits(
            this.pbkdf2Iterations / 10000,
            4
        );

    if (this.encodingMode == 'standard') {
        cryptographicPayload.push(0); // Standard
        cryptographicPayload.push(1); // Mode
    } else {
        cryptographicPayload.push(1); // Compact Mode
    }
    
    for (let bit of saltBits) cryptographicPayload.push(bit);
    for (let bit of iterationsBits) cryptographicPayload.push(bit);

    this.cryptographicPayload = cryptographicPayload;

    let end = new Date().getTime();

    this.debugWrite(id);
    this.debugWrite('    cryptographicPayload:    ', cryptographicPayload);
    this.debugWrite('    iterationsBits:          ', iterationsBits);
    this.debugWrite('    elapsed time (ms):       ', (end - start));

    return this;
}

/**
 *  Async version of {@link Memespeech#computeCryptographicPayload}. Sets the
 *  {@link Memespeech#cryptographicPayload}. Called from 
 *  {@link Memespeech#encrypt} if the Worker is available.
 *  @private
 */
Memespeech.prototype.computeCryptographicPayloadAsync = async function() {
    let id = '(Memespeech.computeCryptographicPayloadAsync)',
        cb = this.workerCallbacks.computeCryptographicPayload,
        start = new Date().getTime()

    return new Promise(function(resolve, reject) {
        cb.resolve = function(cryptographicPayload){
            let end = new Date().getTime();

            this.cryptographicPayload = cryptographicPayload;

            this.debugWrite(id + ' -- Received from Worker!')
            this.debugWrite('    cryptographicPayload: ', cryptographicPayload)
            this.debugWrite('    elapsed time (ms):    ', (end - start))
            resolve(this);
        }.bind(this);

        cb.reject = function(err) {
            console.error(id + ' Worker FAIL: ', err);
            reject(err);
        }.bind(this);

        this.worker.postMessage({
            call: 'computeCryptographicPayload',
            ciphertext: this.ciphertext,
            carrierTextBudget: this.carrierTextBudget,
            pbkdf2Salt: this.pbkdf2Salt,
            pbkdf2Iterations: this.pbkdf2Iterations,
            encodingMode: this.encodingMode
        })
    }.bind(this))
}

/**
 *  Decrypts ciphertext using AES-CBC. The ciphertext and block cipher 
 *  properties for this instance must be populated prior to using this
 *  (Normally this is done by first calling {@link Memespeech#parse} to extract
 *  the Cryptographic Payload out of some Memespeech Text, followed by a call to
 *  {@link Memespeech#initCipherFromPassphrase} to generate the symmetric key).
 *
 *  @returns {String}       The decrypted plaintext
 */
Memespeech.prototype.decrypt = async function() {
    let id = '(Memespeech.decrypt)';

    if (!this.ciphertext) {
        throw id + ': no ciphertext is populated for this instance. Use parse '
            + 'to decode a Cryptographic Payload from some Memespeech Text.';
    }
    if (!this.blockCipherKey || !this.blockCipherIV) {
        throw id + ': cipher is not initialized for this instance. You must '
            + 'first use initCipherFromPassphrase to populate key and IV data.';
    }
    let plaintext,
        start = new Date().getTime();

    try {
        if (this.isNodeJS) {
            let decipher = crypto.createDecipheriv(
                'aes-256-cbc',
                this.blockCipherKey,
                this.blockCipherIV
            );
            let decrypted = decipher.update(this.ciphertext),
                plainBytes = new Uint8Array(Buffer.concat([
                    decrypted, decipher.final()
                ]));

            plaintext = Memespeech.DataUtils.decodePlaintext(
                plainBytes,
                this.encodingMode
            );
        } else {
            let plaintextBytes = new Uint8Array(await crypto.subtle.decrypt(
                { name: 'AES-CBC', iv: this.blockCipherIV },
                this.blockCipherKey,
                this.ciphertext
            ));

            plaintext = Memespeech.DataUtils.decodePlaintext(
                plaintextBytes,
                this.encodingMode
            );
        }
    } catch(err) {
        console.log('err: ', err);
        plaintext = null;
    }

    let end = new Date().getTime();

    if (!plaintext) {
        throw id + ': Could not decrypt! You must either have the wrong '
            + 'passphrase, or maybe there was nothing there to begin with lol';
    }

    this.debugWrite(id);
    this.debugWrite('    plaintext:         ', plaintext);
    this.debugWrite('    elapsed time (ms): ', (end - start));

    return plaintext;
}

/**
 *  Convenience function to both derive the AES key from a passphrase and use it
 *  to decrypt ciphertext. This would be typically done after calling
 *  {@link Memespeech#parse}.
 *
 *  Internally calls {@link Memespeech#initCipherFromPassphrase} followed by
 *  {@link Memespeech#decrypt}, returns a Promise that resolves to the instance.
 *
 *  @param {string} passphrase - The passphrase used for key derivation
 *  @returns {String}            Decrypted plaintext
 */
Memespeech.prototype.passphraseDecrypt = async function(passphrase) {
    await this.initCipherFromPassphrase(passphrase);
    return await this.decrypt();
}

/**
 *  Encode the Cryptographic Payload (if it exists) into the Carrier Text,
 *  transforming the whole thing into Memespeech. If no Cryptographic Payload
 *  exists, then just runs the Memespeech character substitution on the Carrier
 *  Text with nothing hidden inside.
 *  @returns {String}
 */
Memespeech.prototype.toString = function() {
    let id = '(Memespeech.toString)',
        start = new Date().getTime(),
        cryptographicPayload = this.cryptographicPayload || [],
        carrierText = this.carrierText,
        matches = this.carrierTextABCMatches;

    if (!carrierText) {
        throw id +': No Carrier Text found. Start over or RTFM or something.';
    }

    let carrierPos = 0,
        ciphertextPos = 0,
        memespeechStr = '';

    for (var i=0; i<matches.length; i++) {
        let match = matches[i],
            matchPos = match.index;

        for (; carrierPos <= matchPos; carrierPos++) {
            if (carrierPos < matchPos) {
                /*
                 * Add non-alphabetical characters verbatim from Carrier Text
                 */
                memespeechStr += carrierText[carrierPos];
            } else {
                let matchedChar = carrierText[carrierPos];
                
                if (ciphertextPos < cryptographicPayload.length) {
                    /*
                     * Iterate over each bit in the Cryptographic Payload and 
                     * replace the matched alphabetical character from the 
                     * Carrier Text with its lowercased or uppercased equivalent
                     * to denote binary zeroes or ones, respectively.
                     */
                    if (cryptographicPayload[ciphertextPos] == 0) {
                        memespeechStr += matchedChar.toLowerCase();
                    } else {
                        memespeechStr += matchedChar.toUpperCase();
                    }
                    ciphertextPos++
                } else {
                    /*
                     * Once we're done looping over the Cryptographic Payload,
                     * replace any remaining alphabetical characters in the 
                     * Carrier Text with randomized capitalization.
                     */
                    if (Math.random() < .5) {
                        memespeechStr += matchedChar.toLowerCase();
                    } else {
                        memespeechStr += matchedChar.toUpperCase();
                    }
                }
            }
        }
    }
    /*
     * Finish by adding verbatim any remaining non-alphabetical characters.
     */
    for (; carrierPos < carrierText.length; carrierPos++) {
        memespeechStr += carrierText[carrierPos];
    }
    let end = new Date().getTime();

    this.debugWrite(id);
    this.debugWrite('    memespeech string: ', memespeechStr);
    this.debugWrite('    elapsed time (ms): ', (end - start));

    return memespeechStr;
}

/**
 *  Async Web Worker version of {@link Memespeech#toString}.
 *  This method is not supported under Node.js.
 *
 *  @returns {String}
 */
Memespeech.prototype.toStringAsync = async function() {
    let id = '(Memespeech.toStringAsync)',
        cb = this.workerCallbacks.toString,
        start = new Date().getTime();

    if (!this.worker)
        throw id + ': Worker not initialized. Pls call initializeWorker first!';

    return new Promise(function(resolve, reject) {
        cb.resolve = function(memespeechStr){
            let end = new Date().getTime();

            this.debugWrite(id + ' -- Received from Worker!')
            this.debugWrite('    memespeech string: ', memespeechStr);
            this.debugWrite('    elapsed time (ms): ', (end - start));

            resolve(memespeechStr);
        }.bind(this);

        cb.reject = function(err) {
            console.error(id + ' Worker FAIL: ', err);
            reject(err);
        }.bind(this);

        this.worker.postMessage({
            call: 'toString',
            cryptographicPayload: this.cryptographicPayload || [],
            carrierText: this.carrierText,
            matches: this.carrierTextABCMatches
        })
    }.bind(this));
}

/**
 *  Decode and parse a Cryptographic Payload out of a block of Memespeech Text.
 *  This will set {@link Memespeech#ciphertext}, {@link Memespeech#pbkdf2Salt},
 *  and {@link Memespeech#pbkdf2Iterations}. With these properties set on the
 *  instance, it's possible to call {@link Memespeech#initCipherFromPassphrase}
 *  to get the original AES-CBC key and use it with {@link Memespeech#decrypt}.
 *
 *  @param {String} memespeechStr - Memespeech Text to parse
 *  @returns {Memespeech}
 */
Memespeech.prototype.parse = function(memespeechStr) {
    let id = '(Memespeech.parse)',
        start = new Date().getTime();

    /*
     * Set the Carrier Text based on the Memespeech String.
     * this also populates the list of alphabetical matches.
     */
    this.setCarrierText(memespeechStr);

    let matches = this.carrierTextABCMatches,
        rawBinary = [],
        ciphertextBits = [],
        cursor = 0,
        doneParsingCiphertextPacket = false;

    /*
     * Translate the alphabetical matches to "binary" based on capitalization.
     * Some of the raw binary data may be extraneous. We will parse it below.
     */
    for (let match of matches) {
        rawBinary.push(match[0].toLowerCase() === match[0] ? 0 : 1);
    }

    /*
     * Helper function to freak out with an exception if we attempt to parse
     * beyond our binary limit. Obviously it's crap data if that happens.
     */
    let _incrementCursor = function(inc) {
        cursor += inc;
        if (cursor >= rawBinary.length) {
            throw id +': Unable to parse Memespeech Text. Likely bad data :(';
        }
    }

    /*
     * Parse the Ciphertext Packet portion of the Cryptographic Payload.
     * This is split up into "chunks" that begin with a length byte, end with
     * an "is last" bit flag, and have binary ciphertext data sandwiched between
     */
    while (doneParsingCiphertextPacket == false) {
        _incrementCursor(8);
        
        /*
         * Parse the length byte
         */
        let lengthBits = rawBinary.slice(cursor - 8, cursor),
            length = Memespeech.DataUtils.bitsToByte(lengthBits);

        _incrementCursor(length);

        /*
         * Slice ciphertext bits according to the length byte
         */
        let ciphertextSliceBits = rawBinary.slice(cursor - length, cursor);
        for (let bit of ciphertextSliceBits) ciphertextBits.push(bit);

        /*
         * Check if we've hit the "is last" chunk flag
         */
        if (rawBinary[cursor] === 1) {
            doneParsingCiphertextPacket = true;
        }
        _incrementCursor(1);
    }

    /*
     * Now parse the Footer Packet, making sure to check this payload uses the
     * default settings.
     */
    if (rawBinary[cursor] == 1) {            // Encoding Mode: 'compact'
        var saltLength = 64;
        _incrementCursor(1);
        this.encodingMode = 'compact';
    } else if (rawBinary[cursor + 1] == 1) { // Encoding Mode: 'standard'
        var saltLength = 256;                
        _incrementCursor(2);
        this.encodingMode = 'standard';
    } else {                                 // Unsupported encoding mode WTF
        throw id+': Cannot parse non-default Ciphertext Payload Footer Packet.';
    }

    /*
     * The next 256 bits forms our PBKDF2 Salt.
     */
    _incrementCursor(saltLength);
    let saltBits = rawBinary.slice(cursor - saltLength, cursor);

    /*
     * The next 16 bits forms our PBKDF2 Iteration Count
     */
    _incrementCursor(4);
    let iterationsBits = rawBinary.slice(cursor - 4, cursor);

    /*
     * Finally convert all this stuff into usable data
     */
    this.ciphertext = Memespeech.DataUtils.bitsToBytes(ciphertextBits);
    this.pbkdf2Salt = Memespeech.DataUtils.bitsToBytes(saltBits);
    this.pbkdf2Iterations = Memespeech.DataUtils.bitsToInt(iterationsBits)*10000
    
    let end = new Date().getTime();

    this.debugWrite(id);
    this.debugWrite('    parsed ciphertext:       ', this.ciphertext);
    this.debugWrite('    parsed pbkdf2Salt:       ', this.pbkdf2Salt);
    this.debugWrite('    parsed pbkdf2Iterations: ', this.pbkdf2Iterations);
    this.debugWrite('    elapsed time (ms): ', (end - start));

    return this;
}

/**
 *  Async Web Worker version of {@link Memespeech#parse}. Sets all the same
 *  instance properties that method does, but uses a separate CPU thread for
 *  MAXIMUM SMOOTHNESS.
 *
 *  This method is not supported under Node.js.
 *
 *  @param {String} memespeechStr - Memespeech Text to parse
 *  @returns {Memespeech}
 */
Memespeech.prototype.parseAsync = async function(memespeechStr) {
    let id = '(Memespeech.parseAsync)',
        cb = this.workerCallbacks.parse,
        start = new Date().getTime();

    if (!this.worker)
        throw id + ': Worker not initialized. Pls call initializeWorker first!';

    await this.setCarrierTextAsync(memespeechStr);

    return new Promise(function(resolve, reject) {
        cb.resolve = function(data){
            let end = new Date().getTime();

            this.ciphertext = data.ciphertext;
            this.pbkdf2Salt = data.pbkdf2Salt;
            this.pbkdf2Iterations = data.pbkdf2Iterations;
            this.encodingMode = data.encodingMode;

            this.debugWrite(id + ' -- Received from Worker!')
            this.debugWrite('    ciphertext:        ', data.ciphertext);
            this.debugWrite('    pbkdf2Salt:        ', data.pbkdf2Salt);
            this.debugWrite('    pbkdf2Iterations:  ', data.pbkdf2Iterations);
            this.debugWrite('    elapsed time (ms): ', (end - start))
            resolve(this);
        }.bind(this);

        cb.reject = function(err) {
            console.error(id + ' Worker FAIL: ', err);
            reject(err);
        }.bind(this);

        this.worker.postMessage({
            call: 'parse',
            matches: this.carrierTextABCMatches
        })
    }.bind(this));
}

/**
 *  Computes the size of the smallest Cryptographic Payload (in bits) for a
 *  plaintext string by converting it to a Uint8Array of bytes, measuring the
 *  length and adding in what we can predict about the AES hasher block output
 *  size and Memespeech binary encoding practices. NOTE: this computes the
 *  _minimum_ possible size. In practice, the payload may be larger if there's
 *  room to spare in the budget, as splitting the ciphertext up into smaller
 *  chunks gives a better aesthetic appearance of randomness.
 *
 *  @param {string} plaintext - A plaintext string to compute payload size for
 *
 *  @returns {Object}           Object with keys for computed minimum
 *      cryptographic payload sizes in both Standard Mode (totalBitsStandard)
 *      and Compact Mode (totalBitsCompact)
 */
Memespeech.prototype.computeMinCryptographicPayloadSize = function(plaintext) {
    let bytesStd = Memespeech.DataUtils.encodePlaintext(plaintext, 'standard'),
        blockOutBytesStd  = (Math.floor(bytesStd.length / 16) + 1) * 16,
        minDivBytesStd    = (Math.floor((blockOutBytesStd*8) / 255) + 1),
        saltSizeStandard  = 32,
        bytesSmall = Memespeech.DataUtils.encodePlaintext(plaintext, 'compact'),
        blockOutBytesSmall= (Math.floor(bytesSmall.length / 16) + 1) * 16,
        minDivBytesSmall  = (Math.floor((blockOutBytesSmall*8) / 255) + 1),
        saltSizeCompact   = 8,
        looseBitsStandard = minDivBytesStd + 2 + 4,
        looseBitsCompact  = minDivBytesSmall + 1 + 4,
        totalBitsStandard = 8 * (
                blockOutBytesStd  
                + minDivBytesStd
                + saltSizeStandard
            ) + looseBitsStandard,
        totalBitsCompact  = 8 * (
                blockOutBytesSmall  
                + minDivBytesSmall
                + saltSizeCompact
            ) + looseBitsCompact;

    return {
        totalBitsStandard: totalBitsStandard,
        totalBitsCompact: totalBitsCompact
    };
}

/**
 *  Computes the budget (in bytes) for AES plaintext within the instance's
 *  {@link Memespeech#carrierText}, in both 'standard' and 'compact' modes.
 *  This allows you to make an *estimated* character counter for text to encrypt
 *  into the Carrier Text (note this must always be displayed as an estimate to
 *  the user since the budget is returned in bytes and some characters are
 *  multibyte). If one or both of the returned budget numbers is negative, then
 *  there's no way to encode *any* ciphertext into the given Encoding Mode(s).
 *
 *  Throws an error if there is no Carrier Text set on this instance.
 *
 *  @returns {Object}           Object with keys for computed byte budget
 *      for AES encryption in both Standard Mode (byteBudgetStandard)
 *      and Compact Mode (byteBudgetCompact)
 */
Memespeech.prototype.computeAESByteBudget = function() {
    if (!this.carrierText)
        throw '(Memespeech.computeAESByteBudget): No Carrier Text is set.';

    // Find Ciphertext Packet budget by subtracting the Footer Packet size
    let bitBudgetStandard = this.carrierTextBudget - 262,
        bitBudgetCompact = this.carrierTextBudget - 69;

    // Divide the Ciphertext Packet budget into 137 bit chunks. Each chunk
    // contains 128 bits of payload data plus a length byte and "last" flag bit.
    // Multiplying the number of chunks by 16 *almost* gives us the total number
    // of available bytes, but we have to subtract one byte (since the AES PKCS7
    // padding will add an additional block if we completely use up the last 16
    // byte block). Also, in the case of Standard Mode we subtract an additional
    // 4 bytes for the "signature" (see {@link Memespeech#encodePlaintext})
    let byteBudgetStandard = Math.floor(bitBudgetStandard / 137) * 16 - 5,
        byteBudgetCompact = Math.floor(bitBudgetCompact / 137) * 16 - 1;

    return {
        byteBudgetStandard: byteBudgetStandard,
        byteBudgetCompact: byteBudgetCompact
    };
}

/**
 *  Enables or disables Debug Mode (verbose log output) for this instance
 *
 *  @param {boolean} active - Whether to enable or disable Debug Mode
 *  @returns {Memespeech}
 */
Memespeech.prototype.setDebugModeActive = function(active) {
    if (active) {
        this.debugMode = true;
        this.debugWrite('(MEMESPEECH DEBUG MODE ACTIVATED LOL)');
    } else {
        this.debugMode = false;
    }
    return this;
};

/**
 *  Writes debug info to the console if {@link Memespeech#debugMode} is enabled.
 *  @param {...mixed} arguments The variables to write to console
 *
 *  @private
 */
Memespeech.prototype.debugWrite = function() {
    if (this.debugMode) console.log.apply(console, arguments);
}

/**
 *  @classdesc
 *  This object offers various data conversion functions for use in crypto
 *  operations. Think of it as a grab bag of useful static methods.
 *
 *  @namespace
 *  @type {Object}
 */
Memespeech.DataUtils = {
    /**
     *  Converts an ASCII or UTF-8 string to a Uint8Array (bytes)
     *
     *  @static
     *  @param {string} str - The input string
     *  @return {Uint8Array} The byte array
     */
    stringToBytes: function(str) {
        return new TextEncoder("utf-8").encode(str)
    },

    /**
     *  Converts a Uint8Array (bytes) to a UTF-8 string
     *
     *  @static
     *  @param {Uint8Array} uint8Array - The byte array
     *  @returns {string}                UTF-8 string
     */
    bytesToString: function(uint8Array) {
        return new TextDecoder("utf-8").decode(uint8Array);
    },

    /**
     *  Encodes a plaintext string to a Uint8Array (bytes)
     *
     *  If encodingMode is 'standard' then we append a "poor man's signature"
     *  to its UTF-8 bytes. We refer to this signature after decryption to 
     *  verify that the correct key was used (I mean what are the chances 4 
     *  specific bytes would come up randomly??)
     *
     *  NOTE: If encodingMode is 'compact' (or anything other than 'standard'),
     *  we don't add any signature data and just return the plaintext encoded as
     *  a byte array. This will lead to ugly results when decrypting with a bad
     *  key (it will be impossible to verify the output), but this is necessary
     *  in order to save space.
     *
     *  @static
     *  @param {string} str - The input string
     *  @param {String} encodingMode - Either 'standard' or 'compact'. In the
     *      latter case there is no signature added whatsoever to save space.
     *  @return {Uint8Array} The byte array
     */
    encodePlaintext: function(plaintext, encodingMode) {
        let u = Array.from(this.stringToBytes(plaintext));
        if (encodingMode == 'standard') {
            u.push(240);
            u.push(159);
            u.push(145);
            u.push(140);
        }
        return Uint8Array.from(u);
    },

    /**
     *  Converts bytes from encryption to a string.
     *
     *  If encodingMode is 'standard' then we check the "poor man's signature"
     *  previously added by {@link Memespeech.DataUtils#encodePlaintext}. If it
     *  checks out we slice the signature off and return the decoded UTF-8
     *  string. Otherwise return null and leave it to the caller to panic.
     *
     *  If encodingMode is 'compact' (or anything other than 'standard') we 
     *  simply convert the bytes to a string with no signature check whatsoever.
     *  This will lead to very ugly results when decrypting with a bad key but 
     *  is sadly necessary to save space.
     *
     *  @static
     *  @param {Uint8Array} u      - The input Uint8Array
     *  @param {String} encodingMode - Either 'standard' or 'compact'.
     *  @return {String|null}   A string on successful decode, or null for fail.
     */
    decodePlaintext: function(u, encodingMode) {
        if (encodingMode != 'standard') {
            return this.bytesToString(Uint8Array.from(u));
        }
        if (   u[u.length - 1] == 140
            && u[u.length - 2] == 145
            && u[u.length - 3] == 159
            && u[u.length - 4] == 240
        ) {
            return this.bytesToString(Uint8Array.from(u.slice(0, u.length-4)));
        }
        return null;
    },

    /**
     *  Converts a Uint8Array (bytes) to an array of "binary bits", where each
     *  "bit" is represented by an integer 0 or 1. More precisely, the integer
     *  bytes of the input Uint8Array are represented by groups of 8 "bits" in
     *  the output array (in big-endian format), in the same order as the input.
     *
     *  @static
     *  @param {Uint8Array|Array} uint8Array - The byte array
     *  @returns {Array}                       An array of "bits"
     *
     *  @example
     *
     *  Memespeech.DataUtils.bytesToBits([69, 150], true);
     *      // returns: [0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0]
     *      //          (69  in binary is 01000101)
     *      //          (150 in binary is 10010110)
     */
    bytesToBits: function(uint8Array) {
        let allBits = [];

        for (let byte of uint8Array) {
            for (let bit of this.byteToBits(byte)) {
                allBits.push(bit);
            }
        }
        return allBits;
    },

    /**
     *  Turn a single integer byte into an array of "binary bits", where each
     *  bit is represented by an integer 0 or 1, returned in big-endian format.
     *
     *  @static
     *  @param {Uint8Array} uint8Array - The byte array
     *  @param {Number} [length=8]     - Integer number of bits to preserve
     *                                   (default 8, can be reduced as needed)
     *  @returns {Array}                 An array of bits
     */
    byteToBits: function(int, length) {
        length || (length = 8)
        let bits = new Array(length);
        for (var i = length - 1; i >= 0; i--) {
            bits[length - 1 - i] = (int >> i) & 1;
        }

        return bits;
    },

    /**
     *  Turn an array of 8 "binary bits", where each bit is represented by an
     *  integer 0 or 1, into the corresponding integer byte value (big-endian).
     *
     *  @static
     *  @param {Array} bits        - Array of "bits"
     *  @returns {Number}            An 8 bit integer byte
     */
    bitsToByte: function(bits) {
        let int = 0;
        for (var i = 7; i >= 0; i--) { int = (bits[7 - i] << i) | int; }

        return int;
    },

    /**
     *  Turns an array of "binary bits", where each bit is represented byt an
     *  integer 0 or 1, into a Uint8Array of integer bytes. This is done by
     *  looping over increments of 8 "bits", and calling
     *  {@link Memespeech#bitsToByte} to get the associated integer.
     *
     *  @param {Array} bits     - Array of "bits" (length must be multiple of 8)
     *  @returns {Uint8Array}
     */
    bitsToBytes: function(bits) {
        let bytes = [];

        for (let i=0; i<bits.length; i+=8) {
            bytes.push(this.bitsToByte(bits.slice(i, i+8)));
        }
        return Uint8Array.from(bytes);
    },

    /**
     *  Turn an integer (up to 32-bits) into an array of "binary bits", where
     *  each bit is represented by an integer 0 or 1 (big-endian format).
     *  Leading zeroes will be added if necessary to pad the output result out
     *  to the next multiple of 8.
     *
     *  For example:
     *
     *      intToBits(255) => [1, 1, 1, 1, 1, 1, 1, 1]
     *      intToBits(256) => [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0]
     *
     *  @static
     *  @param {Uint8Array} int        - The integer
     *  @returns {Array}                 An array of bits
     */
    intToBits: function(int) {
        let binaryStr = int.toString(2),
            binary = [];

        while (binaryStr.length % 8 !== 0) {
            binaryStr = '0' + binaryStr;
        }
        for (var i=0; i<binaryStr.length; i++) {
            binary.push(parseInt(binaryStr[i]));
        }
        return binary;
    },

    /**
     *  Turn an array of "binary bits" (big-endian) into an integer
     *
     *  @param {Array} bits - Array of "bits"
     *  @returns {Number}
     */
    bitsToInt: function(bits) {
        return parseInt(bits.join(''), 2);
    },

    /**
     *  Generates cryptographically secure random bytes. This is only used in
     *  NodeJS since browser JS has a more concise built-in function.
     *
     *  @param {Number} byteLength - number of bytes to generate
     *  @returns {Uint8Array}
     */
    nodeRandomBytes: async function(byteLength) {
        return new Promise((resolve, reject) => {
            crypto.randomFill(new Uint8Array(byteLength), (err, buffer) => {
                resolve(new Uint8Array(buffer));
            })
        });
    }
};